; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: Context is a object that represents drawing surface in FreshLib
;
;  Target OS: Linux
;
;  Dependencies: memory.asm
;
;  Notes:
;_________________________________________________________________________________________

uses libX11

cmClear       =       GXclear
cmAnd         =       GXand
cmAndReverse  =       GXandReverse
cmCopy        =       GXcopy
cmAndInverted =       GXandInverted
cmNoOp        =       GXnoop
cmXor         =       GXxor
cmOr          =       GXor
cmNor         =       GXnor
cmEquiv       =       GXequiv
cmInvert      =       GXinvert
cmXorReverse  =       GXorReverse
cmCopyInverted =      GXcopyInverted
cmOrInverted   =      GXorInverted
cmNand         =      GXnand
cmSet          =      GXset


image_pointer_context = 2

proc AllocateContext, .raster
begin
        push    ebx ecx edx

        stdcall GetMem, sizeof.TContext
        jc      .error
        mov     ebx, eax

        mov     eax, [.raster]
        mov     [ebx+TContext.raster], eax

        cinvoke XCreateGC, [hApplicationDisplay], [ebx+TContext.raster], 0, 0
        mov     [ebx+TContext.handle], eax
        mov     eax, ebx

.exit:
        pop     edx ecx ebx
        clc
        return

.error:
        DebugMsg 'Error allocating context.'
        xor     eax, eax
        pop     edx ecx ebx
        stc
        return
endp




proc ReleaseContext, .context
begin
        push    eax ebx ecx edx
        mov     ebx, [.context]
        cinvoke XFlush, [hApplicationDisplay]
        cinvoke XFreeGC, [hApplicationDisplay], [ebx+TContext.handle]
        cmp     [ebx+TContext.clipregion], 0
        je      @f
        cinvoke XDestroyRegion, [ebx+TContext.clipregion]
@@:
        stdcall FreeMem, ebx

        clc
        pop     edx ecx ebx eax
        return
endp



proc SetLineStyle, .context, .pLineStyle
.color dd ?
begin
        push    eax ebx ecx edx esi

        mov     ebx, [.context]
        mov     esi, [.pLineStyle]

        mov     eax, [esi+TLineStyle.flags]
        mov     ecx, eax
        shr     ecx, 8
        and     eax, $ff        ; cap_style
        and     ecx, $ff        ; join_style

        mov     edx, LineSolid
        cmp     [esi+TLineStyle.DashCount], 0
        je      @f
        mov     edx, LineDoubleDash
@@:
        cinvoke XSetLineAttributes, [hApplicationDisplay], [ebx+TContext.handle], [esi+TLineStyle.width], edx, eax, ecx
        cinvoke XSetForeground,     [hApplicationDisplay], [ebx+TContext.handle], [esi+TLineStyle.color]

        cmp     [esi+TLineStyle.DashCount], 0
        je      .exit

        lea     ecx, [esi+TLineStyle.DashItems]
        cinvoke XSetDashes, [hApplicationDisplay], [ebx+TContext.handle], 0, ecx, [esi+TLineStyle.DashCount]
        cinvoke XSetBackground, [hApplicationDisplay], [ebx+TContext.handle], [esi+TLineStyle.clSpace]

.exit:
        pop     esi edx ecx ebx eax
        return
endp


proc  SetDrawMode, .context, .mode
begin
        push    eax ecx edx
        mov     eax, [.context]
        cinvoke XSetFunction, [hApplicationDisplay], [eax+TContext.handle], [.mode]
        pop     edx ecx eax
        return
endp





proc SetClipRectangle, .context, .pBounds
.rect XRectangle
begin
        push    eax ecx edx esi

        mov     esi, [.context]

        cmp     [esi+TContext.clipregion], 0
        je      @f
        cinvoke XDestroyRegion, [esi+TContext.clipregion]
@@:
        cinvoke XCreateRegion
        mov     [esi+TContext.clipregion], eax

        mov     edx, [.pBounds]
        test    edx, edx
        jz      .remove

        mov     eax, [edx+TBounds.x]
        mov     [.rect.x], ax

        mov     eax, [edx+TBounds.y]
        mov     [.rect.y], ax

        mov     eax, [edx+TBounds.width]
        mov     [.rect.width], ax

        mov     eax, [edx+TBounds.height]
        mov     [.rect.height], ax

        lea     edx, [.rect]
        cinvoke XUnionRectWithRegion, edx, [esi+TContext.clipregion], [esi+TContext.clipregion]
        cinvoke XSetRegion, [hApplicationDisplay], [esi+TContext.handle], [esi+TContext.clipregion]

.exit:
        pop     esi edx ecx eax
        return

.remove:
        cinvoke XSetClipMask, [hApplicationDisplay], [esi+TContext.handle], None
        jmp     .exit
endp




proc IsDrawable, .raster
begin

;       cinvoke  XSetErrorHandler,





endp