; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: Memory based images manipulation library.
;
;  Target OS: Linux
;
;  Dependencies: memory.asm
;
;  Notes:
;_________________________________________________________________________________________

; only 16, 24 and 32 bpp are supported.
; It is stupid to support palleted modes and even 24bpp is not good...
proc CreateImage, .width, .height, .bpp
.scanline dd ?
begin
        push    ebx ecx edx esi

        stdcall GetMem, sizeof.TImage
        jc      .finish

        mov     esi, eax

        cmp     [.width], 0
        jg     @f
        mov     [.width], 1
@@:
        cmp     [.height], 0
        jg     @f
        mov     [.height], 1
@@:

        push    [.width] [.height] [.bpp]
        pop     [esi+TImage.bpp] [esi+TImage.height] [esi+TImage.width]

; allocate image data

        mov     ecx, [.bpp]
        mov     eax, [.height]
        shr     ecx, 3
        imul    ecx, [.width]
        add     ecx, 3
        and     ecx, $fffffffc  ; align scanline to dword.
        mov     [.scanline], ecx
        imul    eax, ecx

        stdcall GetMem, eax
        jc      .error_free
        mov     [esi+TImage.pData], eax

; IMPORTANT!
; The depth - 3rd argument MUST be $18 - I tested it on 16bpp and 32bpp screen setting and it works.
; in all other settings XPutImage fails with segmentation fault or with BadMatch error.
; This procedure is tested only with 32bit images - [.bpp]=32.
; This issue needs some more research, but the documentation is pretty obscure and I am not so good in Linux programming.
        mov     eax, [.bpp]
        add     eax, $0f
        and     eax, $f0
        cinvoke XCreateImage, [hApplicationDisplay], 0, $18, ZPixmap, 0, [esi+TImage.pData], [.width], [.height], eax, [.scanline]
        mov     [esi+TImage.raster], eax

        clc
        mov     eax, esi
.finish:
        pop     esi edx ecx ebx
        return

.error_free:
        stdcall FreeMem, esi
        stc
        jmp     .finish
endp




proc DestroyImage, .ptrImage
begin
        push    eax ecx edx esi

        mov     esi, [.ptrImage]
        test    esi, esi
        jz      @f

        cinvoke XDestroyImage, [esi+TImage.raster]
        stdcall FreeMem, esi
@@:
        pop     esi edx ecx eax
        return
endp




proc DrawImage, .context, .pImage, .x, .y
begin
        push    eax ecx edx

        mov     eax, [.pImage]
        test    eax, eax
        jz      .exit

        mov     ecx, [.context]
        cinvoke XPutImage, [hApplicationDisplay], [ecx+TContext.raster], [ecx+TContext.handle], [eax+TImage.raster], 0, 0, [.x], [.y], [eax+TImage.width], [eax+TImage.height]
.exit:
        pop     edx ecx eax
        return
endp

