; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: Text drawing library.
;
;  Target OS: Linux
;
;  Dependencies:
;
;  Notes: Uses XFT library to renter the text.
;_________________________________________________________________________________________

uses libXft

proc __ColorToRender, .color, .rendercolor
begin
        push    eax ecx edx

        mov     edx, [.rendercolor]

        xor     eax, eax
        xor     ecx, ecx

        mov     ah, byte [.color]
        mov     ch, byte [.color+1]

        mov     [edx+_XftColor.color.blue], ax
        mov     [edx+_XftColor.color.green], cx

        mov     ah, byte [.color+2]
        mov     ch, byte [.color+3]
        xor     ch, $ff
        mov     [edx+_XftColor.color.red], ax
        mov     [edx+_XftColor.color.alpha], cx

        pop     edx ecx eax
        return
endp



proc DrawString, .context, .ptrString, .len, .x, .y, .font, .color
  .cltxt   _XftColor
begin
        push    eax ebx ecx edx esi edi

        cmp     [.font], 0
        jne     @f

        mov     eax, [_SystemFont]
        mov     [.font], eax

@@:
        cinvoke XDefaultVisual, [hApplicationDisplay], 0
        mov     ecx, [.context]
        cinvoke XftDrawCreate, [hApplicationDisplay], [ecx+TContext.raster], eax, 0
        mov     ebx, eax

        mov     ecx, [.context]
        cinvoke XftDrawSetClip, ebx, [ecx+TContext.clipregion]

        lea     eax, [.cltxt]
        stdcall __ColorToRender, [.color], eax

        cinvoke XftDrawStringUtf8, ebx, eax, [.font], [.x], [.y], [.ptrString], [.len]
        cinvoke XftDrawDestroy, ebx

        pop     edi esi edx ecx ebx eax
        return
endp




proc DrawStringOpaque, .context, .ptrString, .len, .x, .y, .font, .color, .background
  .cltxt   _XftColor
begin
        push    eax ebx ecx edx esi edi

        cmp     [.font], 0
        jne     @f

        mov     eax, [_SystemFont]
        mov     [.font], eax

@@:
        cinvoke XDefaultVisual, [hApplicationDisplay], 0
        mov     ecx, [.context]
        cinvoke XftDrawCreate, [hApplicationDisplay], [ecx+TContext.raster], eax, 0
        mov     ebx, eax

        mov     ecx, [.context]
        cinvoke XftDrawSetClip, ebx, [ecx+TContext.clipregion]

        lea     eax, [.cltxt]
        stdcall __ColorToRender, [.color], eax

        cinvoke XftDrawStringUtf8, ebx, eax, [.font], [.x], [.y], [.ptrString], [.len]
        cinvoke XftDrawDestroy, ebx

        pop     edi esi edx ecx ebx eax
        return
endp



proc DrawColoredString, .context, .ptrString, .pCharAttr, .str_len_bytes, .x, .y, .char_offs, .fontwidth, .fontheight
.xbeg   dd ?
.cltxt  _XftColor
.char   dd ?
.height dd ?
.offs   dd ?
.skip   dd ?
begin
        push    eax ebx ecx edx esi edi

        mov     esi, [.pCharAttr]

        mov     eax, [.x]
        mov     [.xbeg], eax

        cinvoke XDefaultVisual, [hApplicationDisplay], 0
        mov     ecx, [.context]
        cinvoke XftDrawCreate, [hApplicationDisplay], [ecx+TContext.raster], eax, 0
        mov     ebx, eax

        mov     ecx, [.context]
        cinvoke XftDrawSetClip, ebx, [ecx+TContext.clipregion]

        lea     edi, [.cltxt]

        stdcall GetTextBounds, [.context], __HeightProbeString, __HeightProbeString.length, [esi+TCharAttr.font]
        mov     [.height], edx

        stdcall GetTextOffset, [.context], __HeightProbeString, __HeightProbeString.length, [esi+TCharAttr.font]
        mov     [.offs], edx

.skip_offset:
        mov     ecx, [.char_offs]
        mov     [.skip], ecx

.drawloop:
        cmp     [.str_len_bytes], 0
        jle     .enddraw

        mov     ecx, [.ptrString]
        stdcall DecodeUtf8, [ecx]
        mov     [.char], edx

        cmp     eax, $01        ; word wrap character
        je      .wrapline

        cmp     [.skip], 0
        jg      .drawok

;        mov     ecx, [.y]
;        sub     ecx, [.offs]
;        sub     ecx, 2
;        stdcall DrawFillRect, [.context], [.x], ecx, [.fontwidth], [.height], [esi+TCharAttr.backbround]

        stdcall __ColorToRender, [esi+TCharAttr.color], edi
        cinvoke XftDrawStringUtf8, ebx, edi, [esi+TCharAttr.font], [.x], [.y], [.ptrString], [.char]

        mov     eax, [.fontwidth]
        add     [.x], eax
;        sub     [.width], eax
;        js      .enddraw

.drawok:
        dec     [.skip]
        mov     eax, [.char]
        add     [.ptrString], eax
        sub     [.str_len_bytes], eax

        add     esi, sizeof.TCharAttr
        jne     .drawloop

.enddraw:
        cinvoke XftDrawDestroy, ebx
        pop     edi esi edx ecx ebx eax
        return

.wrapline:
        mov     eax, [.fontheight]
        add     [.y], eax
        mov     eax, [.xbeg]
        mov     [.x], eax

        mov     eax, [.char]
        add     [.ptrString], eax
        sub     [.str_len_bytes], eax

        add     esi, sizeof.TCharAttr
        jmp     .skip_offset

endp







; returns:
;       eax - width of the string in pixels.
;       edx - heigth of the string in pixels.


proc GetTextBounds, .context, .ptrString, .len, .font
.extents _XGlyphInfo
begin
        push    esi ecx

        cmp     [.font], 0
        jne     @f

        mov     eax, [_SystemFont]
        mov     [.font], eax

@@:
        lea     esi, [.extents]
        cinvoke   XftTextExtentsUtf8, [hApplicationDisplay], [.font], [.ptrString], [.len], esi

        movzx   eax, [esi+_XGlyphInfo.xOff]
        movzx   edx, [esi+_XGlyphInfo.height]
        add     edx, 2

        pop     ecx esi
        return
endp


; returns:
;       eax - x offset of the baseline of the string.
;       edx - y offset of the baseline of the string.

proc GetTextOffset, .context, .ptrString, .len, .font
.extents _XGlyphInfo
begin
        push    esi ecx

        cmp     [.font], 0
        jne     @f

        mov     eax, [_SystemFont]
        mov     [.font], eax

@@:
        lea     esi, [.extents]
        cinvoke   XftTextExtentsUtf8, [hApplicationDisplay], [.font], [.ptrString], [.len], esi

        movzx   eax, [esi+_XGlyphInfo.x]
        movzx   edx, [esi+_XGlyphInfo.y]
        add     edx, 2

        pop     ecx esi
        return
endp
