; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: Provides raster bitmap, to be used for double buffer painting.
;
;  Target OS: Win32
;
;  Dependencies:
;
;  Notes:
;_________________________________________________________________________________________



proc CreateBackBuffer, .hwin, .width, .height
begin
        push    ebx ecx edx

        stdcall GetMem, sizeof.TBackBuffer
        mov     ebx, eax

        mov     ecx, [.width]
        mov     edx, [.height]

        test    ecx, ecx
        jns     @f
        xor     ecx, ecx
@@:
        test    edx, edx
        jns     @f
        xor     edx, edx
@@:
        mov     [ebx+TBackBuffer.width], ecx
        mov     [ebx+TBackBuffer.height], edx

        invoke  GetDC, [.hwin]
        push    eax
        invoke  CreateCompatibleBitmap, eax, [ebx+TBackBuffer.width], [ebx+TBackBuffer.height]
        mov     [ebx+TBackBuffer.raster], eax

        invoke  ReleaseDC, [.hwin] ; from the stack

        mov     eax, ebx
        pop     edx ecx ebx
        return
endp



proc DestroyBackBuffer, .pBackBuffer
begin
        push    eax ecx edx
        mov     eax, [.pBackBuffer]
        push    eax
        invoke  DeleteObject, [eax+TBackBuffer.raster]
        stdcall FreeMem ; from the stack
        pop     edx ecx eax
        return
endp



proc DrawBackBuffer, .context, .pBackBuffer, .x, .y
begin
        push    eax ecx edx esi edi
        mov     esi, [.pBackBuffer]

        stdcall AllocateContext, [esi+TBackBuffer.raster]
        jc      .exit
        mov     edi, eax

        mov     ecx, [.context]
        invoke  BitBlt, [ecx+TContext.handle], [.x], [.y], [esi+TBackBuffer.width], [esi+TBackBuffer.height], [edi+TContext.handle], 0, 0, SRCCOPY

        stdcall ReleaseContext, edi

.exit:
        pop     edi esi edx ecx eax
        return


endp






proc BackBufferToImage, .context, .backbuffer, .bpp
.bi BITMAPINFOHEADER
begin
        pushad

        mov     esi, [.backbuffer]
        mov     ebx, [.context]

        stdcall CreateImage, [esi+TBackBuffer.width], [esi+TBackBuffer.height], [.bpp]
        mov     edi, eax

        mov     [.bi.biSize], 40
        mov     ecx, [esi+TBackBuffer.width]
        mov     edx, [esi+TBackBuffer.height]
        mov     [.bi.biWidth], ecx
        mov     [.bi.biHeight], edx
        mov     [.bi.biPlanes], 1
        mov     ecx, [.bpp]
        mov     [.bi.biBitCount], cx
        mov     [.bi.biCompression], 0

        lea     eax, [.bi]
        invoke  GetDIBits, [ebx+TContext.handle], [esi+TBackBuffer.raster], 0, [esi+TBackBuffer.height], [edi+TImage.pData], eax, DIB_RGB_COLORS

        mov     [esp+4*regEAX], edi
        popad
        return
endp