; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: TGrid object class
;
;  Target OS: Any
;
;  Dependencies:
;  Notes:
;_________________________________________________________________________________________
module "Rectangles library"


; Calculates the intersection of two source rectangles, pointed by [.src1] and [.src2]
; The result rectangle is placed in RECT structure pointed by [.dest]
; All three rectangles may point to the same structures.
;
; Returns CF=0 if rectangles intersect and CF=1 if not.
; If the rectangles does not intersect, the [.dest] rectangle will contain all fields = 0
;
; Procedure preserves all registers.

proc RectIntersect, .dest, .src1, .src2
begin
        pushad

        mov     esi, [.src1]
        mov     edi, [.src2]
        mov     ebx, [.dest]

; max left1, left2
        mov     ecx, [esi+RECT.left]
        mov     eax, [edi+RECT.left]
        cmp     ecx, eax
        cmovl   ecx, eax
        mov     [ebx+RECT.left], ecx

; min right1, right2
        mov     edx, [esi+RECT.right]
        mov     eax, [edi+RECT.right]
        cmp     edx, eax
        cmovg   edx, eax
        mov     [ebx+RECT.right], edx

        cmp     ecx, edx
        jge     .null

; max top1, top2
        mov     ecx, [esi+RECT.top]
        mov     eax, [edi+RECT.top]
        cmp     ecx, eax
        cmovl   ecx, eax
        mov     [ebx+RECT.top], ecx

; min bottom1, bottom2
        mov     edx, [esi+RECT.bottom]
        mov     eax, [edi+RECT.bottom]
        cmp     edx, eax
        cmovg   edx, eax
        mov     [ebx+RECT.bottom], edx

        cmp     ecx, edx
        jge     .null

        clc
        popad
        return

.null:
        xor     eax, eax
        mov     [ebx+RECT.left], eax
        mov     [ebx+RECT.top], eax
        mov     [ebx+RECT.right], eax
        mov     [ebx+RECT.bottom], eax
        stc
        popad
        return
endp




proc RectCopy, .dest, .src
begin
        pushad

        mov     esi, [.src]
        mov     edi, [.dest]

        mov     eax, [esi+RECT.left]
        mov     ecx, [esi+RECT.top]
        mov     edx, [esi+RECT.right]
        mov     ebx, [esi+RECT.bottom]

        mov     [edi+RECT.left], eax
        mov     [edi+RECT.top], ecx
        mov     [edi+RECT.right], edx
        mov     [edi+RECT.bottom], ebx

        popad
        return
endp



proc RectSort, .src
begin
        push    eax ecx esi

        mov     esi, [.src]

        mov     eax, [esi+RECT.left]
        mov     ecx, [esi+RECT.right]
        cmp     ecx, eax
        jge     @f
        mov     [esi+RECT.right], eax
        mov     [esi+RECT.left],ecx
@@:
        mov     eax, [esi+RECT.top]
        mov     ecx, [esi+RECT.bottom]
        cmp     ecx, eax
        jg      @f
        mov     [esi+RECT.top],ecx
        mov     [esi+RECT.bottom], eax
@@:
        pop     esi ecx eax
        return
endp



proc PointInRect, .pRect, .x, .y
begin
        push    eax esi

        mov     esi, [.pRect]
        mov     eax, [.x]

        cmp     eax, [esi+RECT.left]
        jl      .no
        cmp     eax, [esi+RECT.right]
        jge     .no

        mov     eax, [.y]
        cmp     eax, [esi+RECT.top]
        jl      .no
        cmp     eax, [esi+RECT.bottom]
        jg      .no

        clc
        pop     esi eax
        return

.no:
        stc
        pop     esi eax
        return
endp


; converts RECT to TBounds structure.

proc RectToBounds, .ptrBounds, .ptrRect
begin
        pushad

        mov     esi, [.ptrRect]
        mov     edi, [.ptrBounds]

        mov     ebx, [esi+RECT.left]
        mov     edx, [esi+RECT.top]
        mov     eax, [esi+RECT.right]
        mov     ecx, [esi+RECT.bottom]

        sub     eax, ebx
        sub     ecx, edx

        mov     [edi+TBounds.x], ebx
        mov     [edi+TBounds.y], edx
        mov     [edi+TBounds.width], eax
        mov     [edi+TBounds.height], ecx

        popad
        return
endp



endmodule
