; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: Text drawing library.
;
;  Target OS: Any
;
;  Dependencies: StrLib
;
;  Notes: This file contains OS independent part of the library and includes respective
;         OS dependent files.
;_________________________________________________________________________________________
module "Text library"

include "%TargetOS%/text.asm"


macro min a, b, use {
        sub     b, a
        sbb     use, use
        and     use, b
        add     a, use
}

macro max a, b, use {
        sub     b, a
        sbb     use, use
        not     use
        and     use, b
        add     a, use
}


struct TTextExtent
  .width  dd ?
  .height dd ?
  .OffsX  dd ?
  .OffsY  dd ?
ends


struct TCharAttr
  .font       dd ?
  .color      dd ?
  .background dd ?
  .flags      dd ?      ; for free use of the client.
ends


; dtfXXX means "draw text flag"

; horizontal align
dtfAlignLeft    = 0
dtfAlignRight   = 1
dtfAlignCenter  = 2
dtfAlignJustify = 3

; vertical align
dtfAlignTop    = 0 shl 2
dtfAlignBottom = 1 shl 2
dtfAlignMiddle = 2 shl 2

; Text layout
dtfCRLF         = $0100
dtfWordWrap     = $0200
dtfTableTabs    = $0400


struct _TTextChunk
  .y      dd ?
  .width  dd ?
  .height dd ?
  .len    dd ?
  .ptr    dd ?
ends

__HeightProbeString text '18ilj.Xygl[_|"`g'
__NumberProbeString text '8888888888888888'

; returns: eax - mean width
; returns: edx - height
; returns: ecx - descent
proc FontGetCharSize, .raster, .font, .probe
begin
        push    ebx

        stdcall AllocateContext, [.raster]
        mov     ebx, eax

        stdcall GetTextOffset, ebx, [.probe], 16, [.font]
        mov     ecx, edx

;        DebugMsg "Text offset y:"
;        OutputRegister regECX, 10

        stdcall GetTextBounds, ebx, [.probe], 16, [.font]
        stdcall ReleaseContext, ebx

        sub     ecx, edx
        shr     eax, 4
        neg     ecx
        pop     ebx
        return
endp




proc AdjustCountUtf8
begin
        push    esi
        lea     esi, [esi+ebx]
        stdcall ScanBackUtf8
        mov     ebx, esi
        pop     esi
        sub     ebx, esi
        return
endp


proc DrawTextBox, .context, .text, .bounds, .flags, .font, .color
.len   dd ?
.count dd ?
.height dd ?
.y      dd ?

.low  dd  ?
.high dd  ?

.linew     dd ?
.lineh     dd ?

.DrawProc dd ?
begin
        pushad

        mov     ecx, DrawString

        mov     eax, [.flags]
        and     eax, $03
        cmp     eax, dtfAlignJustify
        jne     @f
        mov     ecx, DrawStringJustified
@@:
        mov     [.DrawProc], ecx

        stdcall StrLen, [.text]
        mov     [.len], eax

        stdcall StrPtr, [.text]
        mov     esi, eax

        mov     edi, [.bounds]
        mov     eax, [edi+TBounds.height]
;        mov     ecx, [edi+TBounds.y]
        mov     [.height], eax
        mov     [.y], 0 ;ecx

; slice the string on chunks, depending on flags, rectangle width and the string content.
; push _TTextChunk for every chunk in stack.
        mov     [.count], 0
.slice:
        mov     ecx, [.len]

        test    [.flags], dtfCRLF
        jz      .cropit

        xor     ecx, ecx

.scanCRLF:
        cmp     ecx, [.len]
        jae     .cropit

        mov     al, [esi+ecx]

        cmp     al, $0d
        je      .cropit
        cmp     al, $0a
        je      .cropit
        inc     ecx
        jmp     .scanCRLF

.cropit:
        mov     ebx, ecx  ; high limit

        stdcall AdjustCountUtf8
        stdcall GetTextBounds, [.context], esi, ebx, [.font]

        cmp     edx, [.height]
        jg      .drawchunks

        mov     [.lineh], edx

        cmp     eax, [edi+TBounds.width]
        jle     .itfits

        mov     [.high], ebx
        mov     [.low], 0

.croploop:
        mov     ebx, [.low]
        add     ebx, [.high]
        shr     ebx, 1

        stdcall AdjustCountUtf8
        stdcall GetTextBounds, [.context], esi, ebx, [.font]
        cmp     eax, [edi+TBounds.width]
        je      .itfits

        jg      .above

        cmp     ebx, [.low]
        je      .itfits

        mov     [.low], ebx
        jmp     .croploop

.above:
        mov     [.high], ebx
        jmp     .croploop

.itfits: ; here ebx contains the len of string that fits in the rectangle.
        mov     ecx, ebx
        mov     [.linew], eax

        test    [.flags], dtfWordWrap
        jz      .pushit

; search the last space or tab that separates words:
        mov     edx, ecx
        cmp     byte [esi+edx], $0d
        je      .found
        cmp     byte [esi+edx], $0a
        je      .found

.searchword:
        cmp     byte [esi+edx], 0
        je      .found
        cmp     byte [esi+edx], ' '
        je      .found
        cmp     byte [esi+edx], $09
        je      .found
        cmp     byte [esi+edx], $0d
        je      .found
        cmp     byte [esi+edx], $0a
        je      .found

        dec     edx
        jnz     .searchword

        mov     edx, ecx

.found:
        push    ebx
        mov     ebx, edx
        stdcall AdjustCountUtf8
        stdcall GetTextBounds, [.context], esi, ebx, [.font]
        mov     [.linew], eax
        mov     [.lineh], edx
        mov     ecx, ebx
        pop     ebx

.pushit:
        push    esi             ; pointer to the string.
        push    ecx             ; length of the string.
        push    [.lineh]        ; height of the line.
        push    [.linew]        ; width of the line.
        push    [.y]            ;

        inc     [.count]

        mov     eax, [.lineh]
        add     [.y], eax
        sub     [.height], eax

        lea     esi, [esi+ecx]
        sub     [.len], ecx
        jz      .drawchunks

        xor     ecx, ecx

; search for the whitespace that have to be skip
.skiploop:
        mov     al, [esi]
        cmp     al, $0d
        je      .skipcrlf
        cmp     al, $0a
        je      .skipcrlf
        cmp     al, $20
        je      .skipspace
        cmp     al, $09
        je      .skipspace

        jmp     .slice

.skipspace:
        inc     esi
        dec     [.len]
        jnz     .skiploop
        jmp     .drawchunks

.skipcrlf:
        inc     esi
        dec     [.len]
        xor     al, $0d xor $0a
        cmp     [esi], al
        jne     .slice
        inc     esi
        dec     [.len]
        jnz     .slice

; pop the _TTextChunk structures and draw, depending on flags.
.drawchunks:
        mov     ebx, esp
        stdcall GetTextOffset, [.context], [ebx+_TTextChunk.ptr], [ebx+_TTextChunk.len], [.font]
        mov     eax, [ebx+_TTextChunk.height]
        sub     eax, edx        ; it should be the line y descent
        test    [.flags], dtfAlignBottom
        jnz     @f
        sub     [.y], eax
@@:
        mov     edx, [.flags]
        and     edx, dtfAlignBottom or dtfAlignMiddle

        mov     ecx, [edi+TBounds.height]
        sub     ecx, [.y]

        cmp     edx, dtfAlignBottom
        je      .yalignok

        sar     ecx, 1
        cmp     edx, dtfAlignMiddle
        je      .yalignok

        xor     ecx, ecx

.yalignok:
.drawloop:
        cmp     [.count], 0
        je      .finish

        mov     ebx, esp

        mov     eax, [edi+TBounds.width]
        sub     eax, [ebx+_TTextChunk.width]

        mov     edx, [.flags]
        and     edx, $03

        cmp     edx, dtfAlignRight
        je      .hready

        shr     eax, 1
        cmp     edx, dtfAlignCenter
        je      .hready

        xor     eax, eax

.hready:
        add     eax, [edi+TBounds.x]

        push    eax
        stdcall GetTextOffset, [.context], [ebx+_TTextChunk.ptr], [ebx+_TTextChunk.len], [.font]
        add     edx, ecx
        add     edx, [ebx+_TTextChunk.y]
        add     edx, [edi+TBounds.y]
        pop     eax

        stdcall [.DrawProc], [.context], [ebx+_TTextChunk.ptr], [ebx+_TTextChunk.len], eax, edx, [.font], [.color]

; draws lines on every row of text. For debug purposes.
iglobal
  if used __UnderlineStyle
    __UnderlineStyle LineStyle 0, $808080
  end if
endg

if defined options.DebugMode & options.DebugMode
        stdcall SetLineStyle, [.context], __UnderlineStyle
        stdcall DrawLine, [.context], [edi+TBounds.x], edx, [edi+TBounds.width], edx
end if
; end debug code


        add     esp, sizeof._TTextChunk
        dec     [.count]
        jmp     .drawloop

.finish:
        popad
        return
endp




proc DrawStringJustified, .raster, .ptrString, .len, .x, .y, .font, .color
.XX    dd ?
.rem   dd ?
.crem  dd ?
.endptr dd ?
begin
        pushad

        mov     esi, [.ptrString]
        mov     eax, [.len]
        add     eax, esi
        mov     [.endptr], eax

        stdcall StrLenUtf8, esi, [.len]
        mov     ecx, eax
        dec     ecx
        jg      @f
        mov     ecx, 1
@@:

        stdcall GetTextBounds, [.raster], esi, [.len], [.font]

        sub     eax, [edi+TBounds.width]
        neg     eax

        cdq
        div    ecx

        cmp     eax, 4
        jge     .cantjustify    ; the needed space is too big for justify, so just draw it left aligned

        mov     ecx, eax
        mov     [.rem], edx
        mov     [.crem], edx

        mov     ebx, [.len]
        push    [.x]
        pop     [.XX]

.drawloop:
        cmp     esi, [.endptr]
        jae     .finish

        stdcall DecodeUtf8, [esi]
        jc      .finish

        push    edx
        stdcall DrawString, [.raster], esi, edx, [.XX], [.y], [.font], [.color]
        stdcall GetTextBounds, [.raster], esi, edx, [.font]
        pop     edx
        add     esi, edx
        add     [.XX], eax
        add     [.XX], ecx

        cmp     [.crem], 0
        je      .drawloop

        sub     ebx, [.rem]
        jge     .drawloop

        inc     [.XX]
        dec     [.crem]
        mov     ebx, [.len]
        sub     ebx, [.rem]
        jmp     .drawloop

.finish:
        popad
        return

.cantjustify:
        stdcall DrawString, [.raster], [.ptrString], [.len], [.x], [.y], [.font], [.color]
        jmp     .finish
endp



endmodule
