; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: Window management OS interface functions.
;
;  Target OS: Linux
;
;  Dependencies:
;
;  Notes:
;_________________________________________________________________________________________


proc _CreateNullWindow
.attr XSetWindowAttributes
begin
        push    ecx edx esi

        lea     eax, [.attr]
        mov     [eax+XSetWindowAttributes.override_redirect], 1
        mov     [eax+XSetWindowAttributes.backing_store], NotUseful
        mov     [eax+XSetWindowAttributes.background_pixmap], None
        cinvoke XCreateWindow, [hApplicationDisplay], [hRootWindow], 0, 0, 1, 1, 0, CopyFromParent, InputOutput, CopyFromParent,  CWBackingStore or CWOverrideRedirect or CWBackPixmap, eax
        mov     esi, eax
        test    eax, eax
        jz      .exit

        cinvoke XSelectInput, [hApplicationDisplay], esi, ExposureMask or StructureNotifyMask or FocusChangeMask or   \
                                                          KeyPressMask or KeyReleaseMask or            \
                                                          ButtonPressMask or ButtonReleaseMask or      \
                                                          EnterWindowMask or LeaveWindowMask or        \
                                                          PointerMotionMask

        cinvoke XSetWMProtocols, [hApplicationDisplay], esi, atomWMDelete, 1
        mov     eax, esi
.exit:
        pop     esi edx ecx
        return
endp



;_________________________________________________________________________________________



proc _DestroyWindow, .hwnd
begin
        push    eax ecx edx
        stdcall _SetWindowStruct, [.hwnd], 0
        cinvoke XDestroyWindow, [hApplicationDisplay], [.hwnd]
        pop     edx ecx eax
        return
endp



;_________________________________________________________________________________________



proc _GetParent, .hwnd
.root dd ?
.parent dd ?
.children dd ?
.count    dd ?
begin
        push    ecx edx esi

        lea     esi, [.count]
        lea     edx, [.children]
        lea     ecx, [.parent]
        lea     eax, [.root]
        cinvoke XQueryTree, [hApplicationDisplay], [.hwnd], eax, ecx, edx, esi
        test    eax, eax
        jz      .exit   ; function failed

        cinvoke XFree, [.children]

        mov     eax, [.parent]

.exit:
        pop     esi edx ecx
        return
endp



;_________________________________________________________________________________________


proc _GetChildren, .hwnd
  .root dd ?
  .parent dd ?
  .children dd ?
  .count    dd ?
begin
        push    ebx ecx edx esi


        lea     ebx, [.count]
        lea     edx, [.children]
        lea     ecx, [.parent]
        lea     eax, [.root]
        cinvoke XQueryTree, [hApplicationDisplay], [.hwnd], eax, ecx, edx, ebx
        test    eax, eax
        jz      .exit   ; function failed

        xor     ebx, ebx
        mov     esi, [.children]
        test    esi, esi
        jz      .endchildren

        stdcall CreateArray, sizeof.TWinArrayItem
        mov     ebx, eax
        mov     ecx, [.count]

.loophwnd:
        dec     ecx
        js      .endchildren_free

        stdcall AddArrayItems, ebx, 1
        mov     ebx, edx

        mov     edx, [esi+4*ecx]
        mov     [eax+TWinArrayItem.handle], edx

        jmp     .loophwnd

.endchildren_free:
        cinvoke XFree, esi

.endchildren:
        mov     eax, ebx
.exit:
        pop     esi edx ecx ebx
        return
endp


;_________________________________________________________________________________________



proc _GetVisible, .hwnd
.attr XWindowAttributes
begin
        push    ebx ecx edx

        lea     ebx, [.attr]
        cinvoke XGetWindowAttributes, [hApplicationDisplay], [.hwnd], ebx
        mov     eax, [ebx+XWindowAttributes.map_state]
        test    eax, eax
        jz      @f
        mov     eax, 1
@@:
        pop     edx ecx ebx
        return
endp

;_________________________________________________________________________________________


proc _GetWindowBounds, .hwnd, .pBounds
begin
        push    eax ecx edx

        int3
        pop     edx ecx eax
        return
endp


;_________________________________________________________________________________________


proc _SetWindowBounds, .hwnd, .pBounds
  .wch XWindowChanges
begin
        push    eax ecx edx

        lea     eax, [.wch]
        mov     edx, [.pBounds]

        push    [edx+TBounds.x] [edx+TBounds.y] [edx+TBounds.width] [edx+TBounds.height]
        pop     [.wch.height] [.wch.width] [.wch.y] [.wch.x]

        cmp     [.wch.width], 0
        jnz     @f
        inc     [.wch.width]
@@:
        cmp     [.wch.height], 0
        jnz     @f
        inc     [.wch.height]
@@:
        cinvoke XConfigureWindow, [hApplicationDisplay], [.hwnd], CWX or CWY or CWWidth or CWHeight, eax

        pop     edx ecx eax
        return
endp


;_________________________________________________________________________________________



proc _SetWindowBorder, .hwnd, .brdType
  .attr XSetWindowAttributes
begin
        push    eax ecx edx

        mov     [.attr.override_redirect], 1
        cmp     [.brdType], borderNone
        je      @f
        mov     [.attr.override_redirect], 0
@@:
        lea     eax, [.attr]
        cinvoke XChangeWindowAttributes, [hApplicationDisplay], [.hwnd], CWOverrideRedirect, eax

        pop     edx ecx eax
        return
endp

;_________________________________________________________________________________________


proc _ShowWindow, .hwnd, .flag
begin
        push    eax ecx edx

        mov     ecx, XUnmapWindow
        cmp     [.flag], 0
        je      @f
        mov     ecx, XMapWindow
@@:
        push    [.hwnd] [hApplicationDisplay]
        call    dword [ecx]
        add     esp, 8

        pop     edx ecx eax
        return
endp


;_________________________________________________________________________________________


proc _RefreshWindow, .hwnd
begin
        push    eax ecx edx
        cinvoke XClearArea, [hApplicationDisplay], [.hwnd], 0, 0, 0, 0, TRUE
        cinvoke XFlush, [hApplicationDisplay]
        pop     edx ecx eax
        return
endp


;_________________________________________________________________________________________


proc _SetFocus, .hwnd
begin
        push    eax ecx edx
        cinvoke XSetInputFocus, [hApplicationDisplay], [.hwnd], RevertToParent, CurrentTime
        pop     edx ecx eax
        return
endp

;_________________________________________________________________________________________

proc _AddChild, .hwnd, .child
begin
        push    eax ecx edx
        cinvoke XReparentWindow, [hApplicationDisplay], [.child], [.hwnd], 10, 10
        pop     edx ecx eax
        return
endp

;_________________________________________________________________________________________


proc _SetWindowTextUtf8, .hwnd, .ptrUtf8
begin
        push    eax ecx edx
        cinvoke XStoreName, [hApplicationDisplay], [.hwnd], [.ptrUtf8]
        pop     edx ecx eax
        return
endp



;_________________________________________________________________________________________



proc _GetWindowTextUtf8, .hwnd, .ptrUtf8
begin
        return
endp




;_________________________________________________________________________________________

proc _EnableWindow, .hwnd, .flag
begin
        push    eax ecx edx

        stdcall _GetChildren, [.hwnd]
        test    eax, eax
        jz      .children_ok

        mov     edx, [eax+TArray.count]
.child_loop:
        dec     edx
        js      .end_children

        stdcall _EnableWindow, [eax+TArray.array+sizeof.TWinArrayItem*edx+TWinArrayItem.handle], [.flag]
        jmp     .child_loop

.end_children:
        stdcall FreeMem, eax

.children_ok:
        mov     ecx, ExposureMask or StructureNotifyMask or EnterWindowMask or LeaveWindowMask or PointerMotionMask
        cmp     [.flag], 0
        je      @f
        or      ecx, FocusChangeMask or KeyPressMask or KeyReleaseMask or    \
                     ButtonPressMask or ButtonReleaseMask
@@:
        cinvoke XSelectInput, [hApplicationDisplay], [.hwnd], ecx

        pop     edx ecx eax
        return
endp

;_________________________________________________________________________________________

proc _SetModalTowards, .hwnd, .hwndParent
begin
        push    eax ecx edx
        cinvoke XSetTransientForHint, [hApplicationDisplay], [.hwnd], [.hwndParent]
        pop     edx ecx eax
        return
endp

;_________________________________________________________________________________________

proc _FinalizeModal, .hwnd, .hwndParent
  .flushevent XEvent
begin
        push    eax ecx edx

        DebugMsg "Finalize modal"

; dirty hack, but the button receives LeaveNotify after DestroyNotify
        cinvoke XSync, [hApplicationDisplay], FALSE
.loop_flush:
        lea     eax, [.flushevent]
        cinvoke XCheckMaskEvent, [hApplicationDisplay], LeaveWindowMask, eax
        test    eax, eax
        jnz     .loop_flush

        pop     edx ecx eax
        return
endp


;_________________________________________________________________________________________


win_structure_context = 1


proc _GetWindowStruct, .hwin
  .user_data dd ?
begin
        push    ecx edx
        lea     eax, [.user_data]
        cinvoke XFindContext, [hApplicationDisplay], [.hwin], win_structure_context, eax
        test    eax, eax
        jnz     .error

        mov     eax, [.user_data]
        clc
        pop     edx ecx
        return

.error:
        DebugMsg 'FreshLib: Error find context.'
        mov      edx, [ebp+4]
        OutputRegister regEDX, 16

        xor      eax, eax
        pop      edx ecx
        return
endp




;_________________________________________________________________________________________



proc _SetWindowStruct, .hwin, .value
begin
        push    eax ecx edx
        cinvoke XSaveContext, [hApplicationDisplay], [.hwin], win_structure_context, [.value]
        pop     edx ecx eax
        return
endp






