; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: Main procedure of GUI application library.
;
;  Target OS: Any
;
;  Dependencies:
;
;  Notes: Organize the main message/event loop needed by every GUI engine.
;         This file contains only OS independent part and includes OS dependent files.
;_________________________________________________________________________________________
module "Main library"

uglobal
  if used pApplication
    pApplication dd ?
  end if
endg


proc Run
begin
.mainloop:
        stdcall ProcessSystemEvents
        jc      .terminate

        mov     eax, [pApplication]
        test    eax, eax
        jz      .eventok

        cmp     [eax+TApplication.OnIdle], 0
        je      .eventok

        pushad
        stdcall [eax+TApplication.OnIdle], eax
        popad

.eventok:
        stdcall WaitForSystemEvent
        jmp     .mainloop

.terminate:
        return
endp




proc ShowModal, .pForm, .pParent
.prev dd ?
begin
        push    ecx edx
        mov     ecx, [.pParent]
        mov     edx, [.pForm]

        jecxz   @f
        stdcall _SetModalTowards, [edx+TForm.handle], [ecx+TWindow.handle]
@@:
        stdcall Set, edx, TForm.Visible, TRUE

        jecxz   @f
        stdcall _EnableWindow, [ecx+TWindow.handle], FALSE
@@:

        mov     [edx+TForm.ModalResult], 0

        stdcall Get, edx, TForm.OnClose
        mov     [.prev], eax
        stdcall Set, edx, TForm.OnClose, __ModalOnClose

.modal_loop:
        stdcall ProcessSystemEvents
        jc      .exit_modal

        cmp     [edx+TForm.ModalResult], 0
        jne     .exit_modal

        stdcall Get, edx, TWindow.Visible
        test    eax, eax
        jz      .exit_modal

.vis_ok:
        stdcall WaitForSystemEvent
        jmp     .modal_loop

.exit_modal:
        jecxz   @f
        stdcall _EnableWindow, [ecx+TWindow.handle], TRUE
@@:

        stdcall Set, edx, TForm.OnClose, [.prev]
        stdcall Set, edx, TForm.Visible, FALSE

        jecxz   @f
        mov     ecx, [ecx+TWindow.handle]
@@:
        stdcall _FinalizeModal, [edx+TForm.handle], ecx

        mov     eax, [edx+TForm.ModalResult]
        pop     edx ecx
        return
endp


; prevent destroy of the form from the close button or Alt+F4
proc __ModalOnClose, .self, .reason
begin
        push    eax
        mov     eax, [.self]
        mov     [eax+TForm.ModalResult], mrCancel
        stc
        pop     eax
        return
endp





include '%TargetOS%/Main.asm'

endmodule