; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: Main procedure of GUI application library.
;
;  Target OS: Win32
;
;  Dependencies:
;
;  Notes: Organize the main message/event loop needed by every GUI engine.
;_________________________________________________________________________________________

proc ProcessSystemEvents
.msg MSG
begin
        push    ecx edx esi

.msgloop:
        lea     esi, [.msg]
        xor     eax, eax

        invoke  PeekMessageW, esi, eax, eax, eax, PM_REMOVE
        test    eax, eax
        jz      .exitok

        cmp     [.msg.message], WM_TIMER
        je      .translate                      ; express dispatching of the timer messages.

        cmp     [.msg.message], WM_QUIT
        je      .terminate

        stdcall __ProcessOneSystemEvent, [esi+MSG.hwnd], [esi+MSG.message], [esi+MSG.wParam], [esi+MSG.lParam]
        jnc     .msgloop                ; the message is translated to system event and the event is executed by the object SysEvent handler.

; the message is for window that is not an FreshLib object, so process with standard windows behaviour.
;        invoke  TranslateMessage, esi
.translate:
        invoke  DispatchMessageW, esi
        jmp     .msgloop


.exitok:
        clc
        pop     esi edx ecx
        return


.terminate:
        mov     eax, [.msg.wParam]
        stc
        pop     esi edx ecx
        return
endp



proc WaitForSystemEvent
begin
        push    eax ecx edx
        invoke  WaitMessage
        pop     edx ecx eax
        return
endp




;proc EventsQueued, .pWin
;.msg MSG
;begin
;        push    eax ecx edx
;        mov     eax, [.pWin]
;        mov     ecx, [eax+TWindow.handle]
;        lea     eax, [.msg]
;        invoke  PeekMessage, eax, ecx, 0, 0, PM_NOREMOVE
;        test    eax, eax
;        clc
;        jz      @f
;        stc
;@@:
;        pop     edx ecx eax
;        return
;endp
;

;----------------------------------------------------------------------------------------------------------
; This procedure makes following:
;   1. Takes as arguments one windows message
;   2. Converts this message to FreshLib system event (or not, depending on the message)
;   3. Calls SysEventHandler procedures for the given object (if any) with the created event structure.
;   4. returns CF=0 if the event was properly processed.
;   5. returns CF=1 if the event was not processed.
;
; The event can be not processed in the following cases:
;   1. The window that receives the message is not FreshLib object.
;   2. TObjectClass.procSysEventHandler = 0 for the given class and all parents.
;   3. All procSysEventHandler procedures refuse to process the event (CF=1)
;----------------------------------------------------------------------------------------------------------
uglobal
  if used __LatestPointedObj
    __LatestPointedObj dd ?
  end if
endg


dproc __ProcessOneSystemEvent, .hwnd, .wmsg, .wparam, .lparam
.event  rb 32
begin
        pushad

        stdcall _GetWindowStruct, [.hwnd]
        test    eax, eax
        jz      .not_processed

        mov     esi, eax
        lea     edi, [.event]

        dispatch [.wmsg]

.ondefault:
        invoke  DefWindowProcW, [.hwnd], [.wmsg], [.wparam], [.lparam]
        mov     [esp+4*regEAX], eax
        popad
        clc
        return

.not_processed:
        popad
        stc
        return

;.........................................................................................

.exec_event:
        stdcall ExecEvent, esi, edi

.finish:
        popad
        xor     eax, eax
        clc
        return


;.........................................................................................


oncase WM_MOUSEWHEEL

          mov   eax, [.wparam]
          sar   eax, 16
          mov   ecx, scWheelUp
          test  eax, eax
          jns   @f
          mov   ecx, scWheelDn
          neg   eax
@@:
          mov   [edi+TScrollEvent.event], seScroll
          mov   [edi+TScrollEvent.ScrollBar], scrollY
          mov   [edi+TScrollEvent.ScrollCmd], ecx

          xor   edx, edx
          mov   ecx, 120
          idiv  ecx
          mov   [edi+TScrollEvent.Value], eax

          jmp   .exec_event

;.........................................................................................

oncase WM_WINDOWPOSCHANGED

locals
  .rect RECT
endl
        mov     ebx, [.lparam]

        lea     eax, [edi+TMoveResizeEvent.newX]
        invoke  GetClientRect, [.hwnd], eax

        mov     eax, [ebx+WINDOWPOS.x]
        mov     ecx, [ebx+WINDOWPOS.y]
        mov     [edi+TMoveResizeEvent.newX], eax
        mov     [edi+TMoveResizeEvent.newY], ecx

        mov     [edi+TMoveResizeEvent.event], seMoveResize
        jmp     .exec_event



;.........................................................................................

oncase WM_SYSKEYDOWN
oncase WM_SYSKEYUP
oncase WM_KEYDOWN
oncase WM_KEYUP
locals
  .keyboard rb 256
  .unicode rw 16
  .utf8    rb 16
endl
        lea     ebx, [.keyboard]
        invoke  GetKeyboardState, ebx
        mov     eax, [.lparam]
        mov     ecx, eax
        shr     eax, 16
        and     ecx, $80000000
        and     eax, $ff
        mov     [edi+TKeyboardEvent.scancode], eax
        or      eax, ecx
        lea     edx, [.unicode]
        invoke  ToUnicode, [.wparam], eax, ebx, edx, 16, 0
        xor     ecx, ecx
        test    eax, eax
        jz      .key_translated    ; the key can not be translated.
        js      .key_translated    ; it is a dead key. What we have to do???

        lea     edx, [.utf8]
        xor     ecx, ecx
        mov     [edx], ecx
        mov     [edx+4], ecx
        lea     ecx, [.unicode]
        invoke  WideCharToMultiByte, CP_UTF8, 0, ecx, eax, edx, 8, 0, 0
        xor     ecx, ecx
        test    eax, eax
        jz      .key_translated

        mov     ecx, dword [.utf8]

.key_translated:
        mov     [edi+TKeyboardEvent.key], ecx

        mov     ecx, seKbdKeyPress
        cmp     [.wmsg], WM_KEYDOWN
        je      @f
        cmp     [.wmsg], WM_SYSKEYDOWN
        je      @f
        mov     ecx, seKbdKeyRelease
@@:
        mov     [edi+TKeyboardEvent.event], ecx

        xor     eax, eax
        test    [.keyboard+VK_CONTROL], $80
        jz      @f
        or      eax, maskCtrl
@@:
        test    [.keyboard+VK_SHIFT], $80
        jz      @f
        or      eax, maskShift
@@:
        test    [.keyboard+VK_MENU], $80
        jz      @f
        or      eax, maskAlt
@@:
        test    [.keyboard+VK_SCROLL], $01
        jz      @f
        or      eax, maskScrLk
@@:
        test    [.keyboard+VK_CAPITAL], $01
        jz      @f
        or      eax, maskCapsLock
@@:

;        or      eax, [__MouseButtonMask]
;        OutputRegister regEAX, 16

        mov     [edi+TKeyboardEvent.kbdStatus], eax
        jmp     .exec_event


;.........................................................................................


oncase WM_SETFOCUS
        mov     [edi+TFocusInEvent.event], seFocusIn
        jmp     .exec_event


;.........................................................................................


oncase WM_KILLFOCUS
        mov     [edi+TFocusOutEvent.event], seFocusOut
        jmp     .exec_event


;.........................................................................................


oncase WM_SETCURSOR

        movzx   eax, word [.lparam]
        cmp     eax, HTCLIENT
        jne     .ondefault

        mov     eax, [esi+TWindow._cursor]
        test    eax, $ffffff00
        jnz     .customcursor

        stdcall GetStockCursor, eax

.customcursor:
        stdcall SetMouseCursor, eax
        jmp     .finish


;.........................................................................................


oncase WM_MOUSEMOVE
        cmp     esi, [__LatestPointedObj]
        je      .normal_mouse_move

        mov     [edi+TMouseEnterEvent.event], seMouseLeave
        xchg    esi, [__LatestPointedObj]
        test    esi, esi
        jz      .leaveok

        stdcall ExecEvent, esi, edi

.leaveok:
        mov     esi, [__LatestPointedObj]
        mov     [edi+TMouseEnterEvent.event], seMouseEnter
        stdcall ExecEvent, esi, edi

.normal_mouse_move:
        mov     [edi+TMouseMoveEvent.event], seMouseMove
        movsx   eax, word [.lparam]
        movsx   ecx, word [.lparam+2]
        mov     [edi+TMouseMoveEvent.x], eax
        mov     [edi+TMouseMoveEvent.y], ecx
        jmp     .exec_event



;.........................................................................................

oncase WM_PAINT
locals
  .ps PAINTSTRUCT
  .context TContext
  .caret   dd ?
  .rect2    RECT
endl

        lea     eax, [.rect2]
        invoke  GetUpdateRect, [.hwnd], eax, FALSE
        test    eax, eax
        jz      .endpaint

        lea     eax, [.ps]
        invoke  BeginPaint, [.hwnd], eax

if defined Caret
        mov     [.caret], -1
        cmp     esi, [Caret.pWindow]
        jne     @f

        stdcall CaretShow, FALSE
        mov     [.caret], eax
@@:
end if

        mov     [edi+TPaintEvent.event], sePaint
        push    [.ps.hdc]
        pop     [.context.handle]
        push    [.hwnd]
        pop     [.context.raster]

        lea     eax, [.context]
        mov     [edi+TPaintEvent.context], eax

; ?????????????????????????????
;        push    [.ps.rcPaint.left] [.ps.rcPaint.top]
        push     [.rect2.left] [.rect2.top]
        pop     [edi+TPaintEvent.rect.top] [edi+TPaintEvent.rect.left]
;        push    [.ps.rcPaint.right] [.ps.rcPaint.bottom]
         push    [.rect2.right] [.rect2.bottom]
        pop     [edi+TPaintEvent.rect.bottom] [edi+TPaintEvent.rect.right]

        stdcall ExecEvent, esi, edi

        lea     eax, [.ps]
        invoke  EndPaint, [.hwnd], eax

if defined Caret
        cmp     [.caret], -1
        je      @f
        stdcall CaretShow, [.caret]
@@:
end if

.endpaint:
        pop     edi esi edx ecx ebx eax
        clc
        return


;.........................................................................................


oncase WM_CLOSE
        mov     eax, [pApplication]
        mov     eax, [eax+TApplication.MainWindow]
        cmp     [eax], esi
        jne     @f

DebugMsg "Quit message posted."
        invoke  PostQuitMessage, 0

@@:
        mov     [edi+TCloseEvent.event], seCloseRequest
        mov     [edi+TCloseEvent.reason], cerFromUser
        jmp     .exec_event

oncase WM_DESTROY
        cmp     esi, [__LatestPointedObj]
        jne     .ondefault

        mov     [__LatestPointedObj], 0
        jmp     .ondefault

;.........................................................................................


oncase WM_LBUTTONUP
oncase WM_RBUTTONUP
oncase WM_MBUTTONUP
oncase WM_LBUTTONDOWN
oncase WM_RBUTTONDOWN
oncase WM_MBUTTONDOWN
oncase WM_LBUTTONDBLCLK
oncase WM_RBUTTONDBLCLK
oncase WM_MBUTTONDBLCLK

        stdcall __MouseButton, [.wmsg]
        mov     [edi+TMouseButtonEvent.event], ecx
        mov     [edi+TMouseButtonEvent.Button], eax
        stdcall __KeyStatus, [.wparam]
        mov     [edi+TMouseButtonEvent.kbdStatus], eax

        movsx   eax, word [.lparam]
        movsx   ecx, word [.lparam+2]
        mov     [edi+TMouseButtonEvent.x], eax
        mov     [edi+TMouseButtonEvent.y], ecx


        jmp     .exec_event
enddp



proc __MouseButton, .msg
begin
        mov     eax, [.msg]
        sub     eax, WM_LBUTTONDOWN
        movzx   ecx, byte [__mouse_event_table+eax]
        movzx   eax, byte [__mouse_button_table+eax]
        return
endp


iglobal
  if used __mouse_button_table
  __mouse_button_table db mbLeft,        mbLeft,       mbLeft
                       db mbRight,       mbRight,      mbRight
                       db mbMiddle,      mbMiddle,     mbMiddle
  end if

  if used __mouse_event_table
  __mouse_event_table  db seMouseBtnPress, seMouseBtnRelease, seMouseBtnDblClick
                       db seMouseBtnPress, seMouseBtnRelease, seMouseBtnDblClick
                       db seMouseBtnPress, seMouseBtnRelease, seMouseBtnDblClick
  end if
endg


;.........................................................................................



proc __KeyStatus, .status
begin
        xor     eax, eax
        test    [.status], MK_LBUTTON
        jz      @f
        or      eax, maskBtnLeft
@@:
        test    [.status], MK_RBUTTON
        jz      @f
        or      eax, maskBtnRight
@@:
        test    [.status], MK_MBUTTON
        jz      @f
        or      eax, maskBtnMiddle
@@:
        test    [.status], MK_SHIFT
        jz      @f
        or      eax, maskCtrl
@@:
        test    [.status], MK_CONTROL
        jz      @f
        or      eax, maskCtrl
@@:
        return
endp