; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: Common dialogs procedures library.
;
;  Target OS: Any
;
;  Dependencies:
;
;  Notes:
;_________________________________________________________________________________________
module "Common dialogs library"


iglobal
   getfile _DlgIconError,       '%lib%/gui/images/error.gif'
   getfile _DlgIconInformation, '%lib%/gui/images/information.gif'
   getfile _DlgIconQuestion,    '%lib%/gui/images/question.gif'
   getfile _DlgIconWarning,     '%lib%/gui/images/warning.gif'
   getfile _DlgMaskDialogIcons, '%lib%/gui/images/icon_mask.gif'

   if used _CommonDialogTemplate
    _CommonDialogTemplate:
        ObjTemplate  tfParent or tfEnd, Form, frmCommonDialog, x, 200, y, 100, width, 480, height, 100
           ObjTemplate tfChild, ImageLabel, .imgIcon,  x, 4, y, 4, width, 50, height, 50, Visible, TRUE, ImageAlign, iaCenter or iaMiddle
           ObjTemplate tfChild, Label,  .LblText,   x,58, y, 4, width, 10, height, 10, TextAlign, dtfAlignLeft or dtfAlignMiddle or dtfWordWrap or dtfCRLF, Visible, TRUE

           ObjTemplate tfChild, Button, .btnOK,     x, 0, y, 0, width, 64, height, 24, TextAlign, dtfAlignCenter  or dtfAlignMiddle, Caption, 'OK',     ModalResult, mrOK
           ObjTemplate tfChild, Button, .btnCancel, x, 0, y, 0, width, 64, height, 24, TextAlign, dtfAlignCenter  or dtfAlignMiddle, Caption, 'Cancel', ModalResult, mrCancel
           ObjTemplate tfChild, Button, .btnAbort,  x, 0, y, 0, width, 64, height, 24, TextAlign, dtfAlignCenter  or dtfAlignMiddle, Caption, 'Abort',  ModalResult, mrAbort
           ObjTemplate tfChild, Button, .btnRetry,  x, 0, y, 0, width, 64, height, 24, TextAlign, dtfAlignCenter  or dtfAlignMiddle, Caption, 'Retry',  ModalResult, mrRetry
           ObjTemplate tfChild, Button, .btnIgnore, x, 0, y, 0, width, 64, height, 24, TextAlign, dtfAlignCenter  or dtfAlignMiddle, Caption, 'Ignore', ModalResult, mrIgnore
           ObjTemplate tfChild, Button, .btnYes,    x, 0, y, 0, width, 64, height, 24, TextAlign, dtfAlignCenter  or dtfAlignMiddle, Caption, 'Yes',    ModalResult, mrYes
           ObjTemplate tfChild, Button, .btnNo,     x, 0, y, 0, width, 64, height, 24, TextAlign, dtfAlignCenter  or dtfAlignMiddle, Caption, 'No',     ModalResult, mrNo
           ObjTemplate tfChild, Button, .btnMaybe,  x, 0, y, 0, width, 64, height, 24, TextAlign, dtfAlignCenter  or dtfAlignMiddle, Caption, 'Maybe',  ModalResult, mrMaybe
           ObjTemplate tfEnd,   Button, .btnHelp,   x, 0, y, 0, width, 64, height, 24, TextAlign, dtfAlignCenter  or dtfAlignMiddle, Caption, 'Help',   ModalResult, mrNone
   end if
endg


; buttons flags used:
smbOK     = $01
smbCancel = $02
smbAbort  = $04
smbRetry  = $08
smbIgnore = $10
smbYes    = $20
smbNo     = $40
smbMaybe  = $80
smbHelp   = $100
smbMaxButton = smbHelp

smiError       = 0
smiInformation = 1
smiQuestion    = 2
smiWarning     = 3



proc ShowMessage, .parent, .icon, .hTitle, .hMessage, .maskButtons
.xrow dd ?
.yrow dd ?
.hicon dd ?
.hmask dd ?
begin
        push    ebx ecx edx esi
        stdcall CreateFromTemplate, _CommonDialogTemplate, 0

; First set the position and size of the dialog box, depending on the flags, and the size of the message.



; Set the message text:
        stdcall Set, [frmCommonDialog.LblText], TLabel.Caption, [.hMessage]

; Set the message title:
        stdcall  Set, ebx, TForm.Caption, [.hTitle]

; arrange the buttons
        mov     eax, [.maskButtons]
        mov     ecx, 32
        xor     edx, edx
.count:
        rol     eax, 1
        adc     edx, 0
        loop    .count

        mov     ecx, edx
        imul    edx, 64         ; button width
        jecxz   @f
        dec     ecx
        imul    ecx, 16
@@:
        add     edx, ecx        ; edx contains the width of the button row in pixels.

        stdcall Get, ebx, TForm.width
        cmp     eax, edx
        jge     .widthok

        lea     eax, [edx+32]
        stdcall Set, ebx, TForm.width, eax

.widthok:
        sub     eax, edx
        sar     eax, 1  ; eax is the begin X coordinate of the row
        mov     [.xrow], eax

        stdcall Get, ebx, TForm.height
        sub     eax, 24+8
        mov     [.yrow], eax

        mov     edx, .btnTable
        mov     ecx, $1

.btnLoop:
        test    [.maskButtons], ecx
        jz      @f

        mov     esi, [edx]
        stdcall Set, [esi], TButton.x, [.xrow]
        stdcall Set, [esi], TButton.y, [.yrow]
        stdcall Set, [esi], TButton.Visible, TRUE
        add     [.xrow], 64+16

@@:
        add     edx, 4
        shl     ecx, 1
        cmp     ecx, smbMaxButton
        jbe     .btnLoop

; set the icon
        mov     eax, [.icon]
        cmp     eax, .iconCount
        jb      @f
        xor     eax, eax
@@:
        stdcall CreateImageGIF, [.iconTable+4*eax], [.iconSizes+4*eax]
        OutputRegister regEAX, 16
        mov     [.hicon], eax
        stdcall Set, [frmCommonDialog.imgIcon], TImageLabel.Image, eax

        stdcall CreateImageGIF, _DlgMaskDialogIcons, _DlgMaskDialogIcons.size
        OutputRegister regEAX, 16
        mov     [.hmask], eax
        stdcall Set, [frmCommonDialog.imgIcon], TImageLabel.Mask, eax

; set the text position and size.
        mov     ecx, [.yrow]
        sub     ecx, 8
        stdcall Set, [frmCommonDialog.LblText], TLabel.height, ecx
        stdcall Set, [frmCommonDialog.imgIcon], TImageLabel.height, ecx

        stdcall Get, [frmCommonDialog], TForm.width
        sub     eax, 66
        stdcall Set, [frmCommonDialog.LblText], TLabel.width, eax

        stdcall ShowModal, ebx, [.parent]
        stdcall Destroy, ebx
        stdcall DestroyImage, [.hicon]
        stdcall DestroyImage, [.hmask]

        pop     esi edx ecx ebx
        return

.btnTable dd frmCommonDialog.btnOK,     \
             frmCommonDialog.btnCancel, \
             frmCommonDialog.btnAbort,  \
             frmCommonDialog.btnRetry,  \
             frmCommonDialog.btnIgnore, \
             frmCommonDialog.btnYes,    \
             frmCommonDialog.btnNo,     \
             frmCommonDialog.btnMaybe,  \
             frmCommonDialog.btnHelp

.iconTable dd _DlgIconError, _DlgIconInformation, _DlgIconQuestion, _DlgIconWarning
.iconCount = ($ - .iconTable)/4
.iconSizes dd _DlgIconError.size, _DlgIconInformation.size, _DlgIconQuestion.size, _DlgIconWarning.size

endp


proc InputString, .parent
begin


        return
endp






endmodule