ttText = 0
ttStr = 1
ttRem = 2
ttNum = 3
ttChar = 4

clDefaultText   = 0
clSymbol        = 1
clNumber        = 2
clString        = 3
clComment       = 4
clRegister      = 5
clInstruction   = 6
clDirective     = 7
clPreprocessor  = 8

SyntaxColorCount = 8

macro dps [txt] {
    forward
      if txt eq 0
        db 0
      else
        db txt
        db 0
        dw 0
      end if
  }


  struc synList minLen, maxLen, ptrlist {
    .MinLen  dd minLen
    .MaxLen  dd maxLen
    .ptrList dd ptrlist
  }

  virtual at 0
    synList synList 0, 0, 0
  end virtual


proc fasm_syntax, .ptrLine, .ptrColors
.status      dd ?
.wlen        dd ?
.w           rb 256
.lastword    dd ?
.Quo         db ?

; Sometimes, AsmEdit calls this procedure with wrong ptrLine.
; In these cases GPF appear.

begin
        pushad
        mov     [.status],ttText
        xor     ah,ah
        mov     [.wlen], 0
        mov     esi, [.ptrLine]
        mov     edi, [.ptrColors]
        mov     [.lastword], edi
        mov     ecx,256

.loop:
        lodsb
        cmp     [.status], ttRem
        je      .endif
        cmp     [.status], ttStr
        jne     .ifnotstr
        cmp     al,[.Quo]
        jne     .endif
        mov     [.status], ttText
        jmp     .endif

.ifnotstr:
        cmp     al, ';'
        jne     .notrem
        mov     [.status], ttRem
        mov     ah, clComment
        jmp     .endif

.notrem:
        cmp     al, '"'
        je      .str
        cmp     al, "'"
        jne     .notstr

.str:
        mov     [.status], ttStr
        mov     [.Quo], al
        mov     ah, clString
        jmp     .endif

.notstr:
        call    .CharInSymbols
        jnc     .notsym
        mov     [.status], ttChar
        mov     ah, clSymbol
        call    .MakeWord
        jmp     .endif

.notsym:
        lea     edx,[esi-1]
        cmp     edx, [.ptrLine]
        je      .maybenum

        cmp     [.status], ttChar
        jne     .notnumbeg

.maybenum:
        call    .CharInNum
        jnc     .notnum
        mov      [.status], ttNum
        mov      ah, clNumber
        jmp      .endif

.notnum:
        mov     [.status], ttText
        xor     ah,ah
        mov     [.wlen], 1
        mov     [.w], al
        mov     [.lastword], edi
        jmp     .endif

.notnumbeg:
        cmp     [.status], ttText
        jne     .endif
        xor     ah,ah
        lea     ebx, [.w]
        add     ebx, [.wlen]
        mov     [ebx], al
        inc     [.wlen]

.endif:
        mov     [edi], ah
        inc     edi
        dec     ecx
        jnz     .loop

.endwhile:
        popad
        return
endp



.MakeWord:
        push    edi esi eax ecx
        cmp     [.wlen],0
        je      .end
        lea     esi, [.w]
        push    esi edi
        lea     edi, [.w]
 .new_char:
        lodsb
        cmp     al,0
        je      .ok
        cmp     al,"A"
        jl      @F
        cmp     al,"Z"
        jg      @F
        add     al,20h
@@:
        stosb
        jmp     .new_char
.ok:
        pop     edi esi

        mov     ah, clRegister
        stdcall FindInstr, sSymbols, esi, [.wlen], 2, 0
        jnc     .maybeinstr

        cmp     dl, 10h
        je      .found

        mov     ah, clDirective
        jmp     .found

.maybeinstr:
        mov     ah, clInstruction
        stdcall FindInstr, sInstructions, esi, [.wlen], 3, 1
        jnc     .maybeoperator

        cmp     edx, fasm_instructions - instruction_handler
        jae     .found

        mov     ah, clDirective
        jmp     .found

.maybeoperator:
        mov     ah, clDirective
        stdcall SearchSimple, directive_operators, esi, [.wlen], 1
        jc      .found

        stdcall FindInstr, sDataDirectives, esi, [.wlen], 3, -1
        jc      .found

        mov     ah, clPreprocessor
        stdcall SearchSimple, preprocessor_directives, esi, [.wlen], 2
        jc      .found
        stdcall SearchSimple, macro_directives, esi, [.wlen], 2
        jnc     .notfound

.found:
        mov     ebx, [.wlen]
        mov     edi, [.lastword]

.chgcolor:
        mov     [edi], ah
        inc     edi
        dec     ebx
        jnz     .chgcolor

.notfound:
        and     [.wlen], 0

.end:
        pop     ecx eax esi edi
        retn


.CharInSymbols:
        push    edi ecx

        mov     edi, symbol_characters+1
        movzx   ecx, byte [edi-1]

        repne scasb
        jne   @f
        stc
        pop     ecx edi
        ret
@@:
        clc
        pop     ecx edi
        retn


.CharInNum:
        cmp     al, '$'
        jne     @f
        stc
        retn

@@:
        cmp     al, '0'
        jb      @f
        cmp     al, '9'
        ja      @f
        stc
        retn
@@:
        clc
        retn




proc FindInstr, .ptrArray, .ptrWord, .len, .addbytes, .return
.count dd ?
begin

        push    eax ebx ecx esi edi

        mov     ecx, [.ptrArray]
        mov     edi, [.ptrWord]
        mov     edx, [.len]

        cmp     edx, [ecx+synList.MinLen]
        jb      .notfound

        cmp     edx, [ecx+synList.MaxLen]
        ja      .notfound

        mov     esi, [ecx+synList.ptrList]

        sub     edx, [ecx+synList.MinLen]
        movzx   eax, word [esi+4*edx+2]       ; items count
        mov     [.count], eax
        movzx   ebx, word [esi+4*edx]         ; offset to esi
        add     esi, ebx
        mov     edx,[.len]
        add     edx, [.addbytes]

.loopmain:
        dec     [.count]
        js      .notfound

        xor     ebx, ebx

.loopword:
        mov     al,[esi+ebx]
        cmp     [edi+ebx], al
        ja      .next
        jb      .notfound
        inc     ebx
        cmp     ebx, [.len]
        jne     .loopword

        cmp     [.return], 0
        jl      .found

        lea     eax, [esi+ebx]
        add     eax, [.return]

        movsx   edx, word [eax]

.found:
        stc
        jmp     .end

.next:
        add     esi, edx
        jmp     .loopmain

.notfound:
        clc

.end:
        pop     edi esi ecx ebx eax
        return
endp



proc SearchSimple, .ptrList, .ptrWord, .len, .addbytes
.count dd ?
begin
        push    eax ecx edx esi edi

        mov     esi, [.ptrList]
        mov     edx, [.len]

.loopmain:
        lodsb
        test    al, al
        jz      .notfound

        cmp     al, dl
        jne     .next

        mov     edi, [.ptrWord]
        movzx   ecx, al
        repe    cmpsb
        je      .found

        add     esi, ecx
        add     esi, [.addbytes]
        jmp     .loopmain

.next:
        movzx   eax, al
        lea     esi, [esi+eax]
        add     esi, [.addbytes]
        jmp     .loopmain

.found:
        stc
        jmp     .end

.notfound:
        clc
.end:
        pop     edi esi edx ecx eax
        return
endp





iglobal
  sInstructions   synList 2, 16, instructions
  sSymbols        synList 2, 11, symbols
  sDataDirectives synList 2, 4, data_directives
endg
