; _____________________________________________________________________________________________
;|                                                                                             |
;| Fresh.asm                                                                                   |
;|                                                                                             |
;| This file is the main file of:                                                              |
;|                                                                                             |
;| ..::Fresh::.. - assembler RAD IDE with built-in FASM assembler.                             |
;|                                                                                             |
;| Credits:                                                                                    |
;|   Tomasz Grysztar (aka Privalov) - for exelent work with FASM - the base for Fresh.         |
;|                                                                                             |
;| The people that contributed to the project:                                                 |
;|                                                                                             |
;|    John Found                                                                               |
;|    Fredrik Klasson (aka scientica)                                                          |
;|    Tommy Lillehagen                                                                         |
;|    Mateusz Tymek (aka decard)                                                               |
;|    Victor Loh (aka roticv)                                                                  |
;|    Yunus Sina Gulsen (aka VeSCeRa)                                                          |
;|                                                                                             |
;|    Tomasz Grysztar - FASM compiler and AsmEdit control.                                     |
;|                                                                                             |
;| Contact the authors: http://board.flatassembler.net/                                        |
;| Fresh IDE main site: http://fresh.flatassembler.net/                                        |
;| FASM main site:      http://www.flatassembler.net/                                          |
;|                                                                                             |
;|_____________________________________________________________________________________________|
;

include '%lib%/freshlib.inc'

@BinaryType GUI

uses kernel32, user32, gdi32, comctl32, comdlg32, ole32

LIB_MODE  equ OLD

if TargetOS eq Win32
  stack 0x10020           ; 64K stack for HtmlHelp :/
end if

options.ShowUserMessages = 0
options.ShowSkipped = 0
options.FastEnter = 0
options.CheckArguments = 0
options.ShowSizes = 1
options.DebugMode = 0


struct MONITORINFO
  .cbSize    dd ?
  .rcMonitor RECT
  .rcWork    RECT
  .dwFlags   dd ?
ends



include 'version.inc'

include '../include/libs/msgutils.inc'
include '../include/libs/gui.inc'
include '../include/libs/parents.inc'
include '../include/libs/tform.inc'
include '../include/libs/templates.inc'
include '../include/libs/application.inc'
include '../include/libs/TDataGrid.inc'
include '../include/libs/buttonedit.inc'

include '../include/libs/asmedit/asmedit.inc'
include 'lister.inc'
include 'dumper.inc'
include 'designtime.inc'
include 'virtfiles.inc'
include 'resourceid.inc'
include 'projmanager.inc'
include 'Controls.inc'
include 'editorhost.inc'
include 'msgwindow.inc'
include '../include/libs/splitter.inc'
include 'propgrid.inc'
include 'MainWindow.inc'
include 'fasm/Version.inc'
include 'xref.inc'
include 'LabelsList.inc'
include 'debug.inc'
include 'debugger.inc'
include 'dbgserver.inc'
include 'propeditor.inc'
include 'filemanager.inc'
include 'FormEditor.inc'
include 'SizerClass.inc'


;;<OS INDEPENDENT SOURCES> all included files should resides here
include "%lib%/freshlib.asm"
include "BrowseHistory.asm"

;<OS DEPENDENT SOURCES> - at the end this list should be empty.

  include 'strutils.asm'        ; some Win32 string utils, not to be ported to other OSes.

  include '../include/libs/asmedit/asmedit.asm'
  include '../include/libs/asmedit/syntax.asm'

  include 'errors.asm'

  include 'LabelsList.asm'
  include 'fasm.asm'

  include 'virtfiles.asm'

  include 'LabelProperties.frm'
  include 'LabelsExplorer.frm'
  include 'LabelInfo.frm'

  include '../include/libs/parents.asm'
  include '../include/libs/tform.asm'
  include '../include/libs/templates.asm'
  include '../include/libs/application.asm'

  include '../include/libs/giflib.asm'  ; this is the old version of giflib.

  include 'Controls.asm'
  include 'SizerClass.asm'

  include '../include/libs/TDataGrid.asm'

  include 'MainWindow.asm'

  include 'ActionEvents.asm'
  include 'actNewFile.asm'
  include 'formutils.asm'
  include 'opendialog.asm'

  include 'propeditor.asm'

  include 'projmanager.asm'

  include 'msgwindow.asm'

  include 'palette.asm'

  include '../include/libs/splitter.asm'
  include '../include/libs/buttonedit.asm'
  include 'propgrid.asm'

  include 'progresswnd.frm'

  include 'findreplace.asm'
  include 'editorhost.asm'
  include 'sourceeditor.asm'
  include 'CCList.asm'
  include 'PAHint.asm'
  include 'EmbededHelp.asm'

  include 'fileutils.asm'
  include 'mrulist.asm'

  include 'IDEparams.asm'

  include 'HelpFilesOptions.asm'

  include 'EditPath.frm'
  include 'EditPathsOptions.frm'
  include 'IDEBehaviourOptions.frm'
  include 'IDEOptions.frm'

  include 'formsDT.asm'
  include 'help.asm'

  include 'lister.asm'
  include 'dumper.asm'

  include 'dbgserver.asm'
  include 'debugger.asm'
  include 'debug.asm'

  include 'dialogs.asm'
  include 'FormEditor.asm'

  include 'filemanager.asm'
  include 'watcheditor.frm'


  include '../include/libs/msgutils.asm'
  include '../include/libs/gui.asm'

  include '../include/libs/scrollbars.asm'

  include 'EDITOptions.frm'
  include 'EditorBehaviour.frm'
  include 'EditorColors.frm'

  include 'frmStyleEditor.frm'
  include 'ProjectOptions.frm'
  include 'GraphicEditor.frm'
  include 'templates.frm'
  include 'ExternalOptions.frm'

  include 'HelpFileList.frm'

; Common initialized globals

; Standard class names
  cEditClassName          text  'EDIT'
  cStaticClassName        text  'STATIC'
  cButtonClassName        text  'BUTTON'
  cComboClassName         text  'COMBOBOX'
  cScrollClassName        text  'SCROLLBAR'
  cListboxClassName       text  'LISTBOX'

; some CommCtrl class names
  cTabClassName           text  TABCONTROL_CLASS
  cTreeViewClassName      text  TREEVIEW_CLASS
  cStatusBarClassName     text  STATUS_CLASS
  cProgressbarClassName   text  PROGRESS_CLASS
  cListviewClassName      text  LISTVIEW_CLASS


uglobal
  hInstance       dd 0  ; Instance handle for common use.
  ComCtrlsFlags   INITCOMMONCONTROLSEX
endg



start:
; Main init sequence
        invoke  GetModuleHandleA,0
        mov     [hInstance],eax

; Init ALL common controls
        mov     [ComCtrlsFlags.dwSize], sizeof.INITCOMMONCONTROLSEX
        mov     [ComCtrlsFlags.dwICC], $1fff
        invoke  InitCommonControlsEx, ComCtrlsFlags

; call all "initialize" procedures in the included modules...
        InitializeAll

; Startup windows creation
        stdcall InitApplication, resAppIcon, cMainTitle

; Accelerator table creating
        invoke  CreateAcceleratorTableA, Accelerators, [Accelerators.Count]

        invoke  SetPropA, [hApplication], [propAccelerators], eax
        invoke  SetPropA, [hApplication], [propOnIdle], OnIdle

        stdcall InitMainWindow, [hApplication]  ; Create main window.
        stdcall AutoArrangeWindows

; Create the tool palette
        stdcall  OnReloadVCL, 0, [hMainWindow]

; Command line parameters.
        stdcall  ParseCommandLine

; Main message loop
Run:
        call    ProcessMessages
        jc      .terminate

if defined AppLib
        invoke  SendMessageA, [hApplication], AM_ONIDLE, 0, 0
end if
        invoke  WaitMessage
        jmp     Run

.terminate:
        push    eax
        FinalizeAll
        stdcall TerminateAll  ; exit code from the stack.


@AllDataSection
@AllImportSection

section '.res' resource from 'images\freshnew.res' data readable
