;---------------------------------------------------------------------
; Converts some filename's path (full or relative) in relative
; form, related to current directory.
; Arg:
;   hFileName - handle/ptr of the filename.
; Returns:
;   eax - Handle to new string containing the relative path/filename
;---------------------------------------------------------------------

proc GetRelativePath, .hFileName
.root rb 512
.name rb 512
.res  rb 512
.ptr  dd ?
begin
        push    esi edi ebx

        lea     esi, [.root]
        invoke  GetCurrentDirectoryA, 512, esi
        cmp     byte [esi+eax-1], '\'
        je      @f
        mov     word [esi+eax], '\'
@@:
        stdcall StrPtr, [.hFileName]
        lea     edi, [.name]
        lea     ecx, [.ptr]
        invoke  GetFullPathNameA, eax, 512, edi, ecx
        test    eax, eax
        jz      .returnempty    ; there is invalid or missing filename.

        lea     edx, [.res]

; find first subdir
        mov     cl, ':'
        call    _CmpSubFolder
        jc      .copy                   ; Copy full path, because they are in different drives.

        mov     cl, '\'

        inc     esi
        inc     edi

.subloop:
        call    _CmpSubFolder
        jnc     .subloop

        dec     esi

; count of remaining subfolders.
.count:
        inc     esi
        mov     al, [esi]
        test    al,al
        je      .copy
        cmp     al,cl
        jne     .count
.found:
        mov     dword [edx], '..\'
        add     edx, 3
        test    al,al
        jnz     .count

.copy:
        mov     al, [edi]
        inc     edi
        mov     [edx], al
        inc     edx
        test    al,al
        jnz     .copy

        stdcall StrNew
        lea     edx, [.res]
        stdcall StrCopy, eax, edx

        pop     ebx edi esi
        return

.returnempty:
        stdcall StrNew
        pop     ebx edi esi
        return
endp




proc _CmpSubFolder       ; esi - first, edi - second, cl - end char
begin
        xor     ebx, ebx

.scanloop:
        mov     al, [esi+ebx]
        test    al,al
        jz      .endofpath

        mov     ch, al
        mov     ah, [edi+ebx]
        or      ax, $2020
        cmp     al, ah
        jne     .notequal

        cmp     ch,cl
        je      .eosubdir

        inc     ebx
        jmp     .scanloop

.endofpath:
        cmp     byte [edi+ebx], cl
        jne     .notequal
        inc     edi
        add     edi, ebx
        add     esi, ebx
        xor     ebx, ebx
        stc
        return

.eosubdir:
        add     esi, ebx
        add     edi, ebx
        inc     esi
        inc     edi
        xor     ebx, ebx
        clc
        return

.notequal:
        stc
        return
endp

;----------------------------------------
; Returns an index to the filename
; separated from full path name.
; The original string remains unchanged
;----------------------------------------
proc ExtractFileName, .hPath
begin
        push    ebx ecx edx
        stdcall StrPtr, [.hPath]
        mov     ebx, eax
        mov     edx, eax
        xor     ecx, ecx
.loop:
        mov     cl, [edx]
        inc     edx
        jecxz   .end

        cmp     cl, '\'
        je      .dir
        cmp     cl, '/'
        jne     .loop
.dir:
        mov     eax, edx
        jmp     .loop

.end:
        sub     eax, ebx
        pop     edx ecx ebx
        return
endp





;----------------------------------------------------------
; Create an unique for the project filename of shape:
; <hPrefix>NNN<hSuffix> - where NNN is the mimimal number
; that makes the filename unique.
; Returns:
;   eax - handle to the new created filename.
;   edx - the NNN number.
;----------------------------------------------------------

proc CreateUniqueName, .hPrefix, .hSuffix
.str rb 1024
.num dd ?
.nm  dd ?
.ptr dd ?
begin
        push    esi edi ebx
        mov     [.num], 1
        stdcall StrNew
        mov     [.nm], eax
        lea     ebx, [.str]

; Create name
.findloop:
        stdcall StrCopy, [.nm], [.hPrefix]

        stdcall NumToStr, [.num], ntsUnsigned or ntsDec
        stdcall StrCat, [.nm], eax
        stdcall StrDel, eax
        stdcall StrCat, [.nm], [.hSuffix]

        stdcall StrPtr, [.nm]
        lea     esi, [.ptr]
        invoke  GetFullPathNameA, eax, 1024, ebx, esi

        stdcall IsFileOpen, ebx
        test    eax, eax
        jz      .found

.nextloop:
        inc     [.num]
        jmp     .findloop

.found:
        stdcall StrCopy, [.nm], ebx
        mov     eax, [.nm]
        mov     edx, [.num]

        pop     ebx edi esi
        return
endp



proc HashFilename, .hFileName
.str rb 1024
.ptr dd ?
begin
        push    esi edi ebx
        lea     esi, [.str]

        stdcall StrPtr, [.hFileName]
        lea     ecx, [.ptr]
        invoke  GetFullPathNameA, eax, 1024, esi, ecx
        stdcall StrLCase, esi
        stdcall StrHash, esi
        pop     ebx edi esi
        return
endp




proc ChangeDir, .hFileName
.ptrname dd ?
.str rb 1024
begin
        push    esi
        stdcall StrPtr, [.hFileName]

        lea     ecx, [.ptrname]
        lea     esi, [.str]
        invoke  GetFullPathNameA, eax, 1024, esi, ecx
        test    eax, eax
        jz      .finish_error

        mov     eax, [.ptrname]
        test    eax, eax
        jz      @f
        mov     byte [eax], 0
@@:
        invoke  SetCurrentDirectoryA, esi
        pop     esi
        return

.finish_error:
        stc
        pop     esi
        return
endp