iglobal
  cOpenFileTitle text 'Open file:'
  cOpenProjectTitle text 'Open project:'
  cSaveProjectTitle text 'Save project as:'
  cSaveFileTitle text 'Save file as:'
  cBackSlash text '\'
  ProjFilter:
        db 'Fresh project (*.fpr)', 0, '*.fpr',0
        db 'All files (*.*)', 0, '*.*',0
        db 0
  AllFilter:
        db 'All Fresh files (*.fpr, *.asm, *.inc, *.frm, *.bmp)', 0, '*.fpr;*.asm;*.inc;*.frm;*.bmp',0
        db 'Fresh project (*.fpr)', 0, '*.fpr',0
        db 'Form (*.frm)', 0, '*.frm',0
        db 'Assembler files (*.asm, *.inc)', 0, '*.asm;*.inc', 0
        db 'Bitmaps (*.bmp)', 0, '*.bmp',0
        db 'All files (*.*)', 0, '*.*',0
        db 0
  FileFilter:
        db 'Assembler files (*.asm, *.inc)', 0, '*.asm;*.inc', 0
        db 'All files (*.*)', 0, '*.*',0
        db 0
  FormFilter:
        db 'Form (*.frm)', 0, '*.frm',0
        db 'All files (*.*)', 0, '*.*', 0
        db 0
  GraphicFilter:
        db 'Bitmap (*.bmp)', 0, '*.bmp',0
        db 'All files (*.*)', 0, '*.*', 0
        db 0
  WatchFilter:
        db 'Memory watch table (*.mwt)', 0, '*.mwt',0
        db 'All files (*.*)', 0, '*.*', 0
        db 0
  cOpenDialogError text 'Error show open dialog.'
endg


cFilenameBufferSize = 4096


proc OpenFilesDialog, .ptrTitle, .MultiSelect, .ptrFilter, .ptrCallBack, .hFileName, .lParam
.opn OPENFILENAME
.str rb 1024
.ptr dd ?
.errcnt dd ?
.oldwin dd ?
begin
        mov     [.errcnt], 0

        invoke  GetActiveWindow
        mov     [.oldwin], eax

; Allocate memory for filename buffer.
        stdcall GetMem, cFilenameBufferSize
        mov     edi, eax

        mov     [.opn.lpstrInitialDir], 0

        cmp     [.hFileName], 0
        je      .setstructure

        stdcall StrPtr, [.hFileName]
        lea     ecx, [.str]
        mov     [.opn.lpstrInitialDir], ecx
        lea     edx, [.ptr]
        invoke  GetFullPathNameA, eax, 1024, ecx, edx
        test    eax, eax
        jz      .setstructure

        mov     esi, [.ptr]
        test    esi, esi
        jz      .setstructure

        mov     byte [esi-1], 0

        push    edi

.copyname:
        lodsb
        stosb
        test    al,al
        jnz     .copyname

        pop     edi

.setstructure:
        xor     edx, edx
        mov     [.opn.lStructSize], sizeof.OPENFILENAME
        mov     eax, [hApplication]
        mov     [.opn.hwndOwner], eax
        mov     [.opn.hInstance], edx
        mov     eax, [.ptrFilter]
        mov     [.opn.lpstrFilter], eax
        mov     [.opn.lpstrCustomFilter], edx
        mov     [.opn.nFilterIndex], 1
        mov     [.opn.lpstrFile], edi
        mov     [.opn.nMaxFile], cFilenameBufferSize
        mov     [.opn.lpstrFileTitle], edx
        mov     [.opn.nMaxFileTitle], edx
        mov     eax, [.ptrTitle]
        mov     [.opn.lpstrTitle], eax
        mov     [.opn.Flags], OFN_EXPLORER or           \
                              OFN_HIDEREADONLY or       \
                              OFN_NOCHANGEDIR or        \
                              OFN_FILEMUSTEXIST

        cmp     [.MultiSelect], FALSE
        je      @f
        or      [.opn.Flags], OFN_ALLOWMULTISELECT
@@:
        mov     [.opn.lpstrDefExt], edx
        mov     [.opn.lCustData], edx
        mov     [.opn.lpfnHook], edx

.retry:
        lea     eax, [.opn]
        invoke  GetOpenFileNameA, eax
        push    eax
        invoke  SetActiveWindow, [.oldwin]      ; restore the old active window.

        pop     eax
        test    eax, eax
        jnz     @f

; Error open dialog or simply cancel...
        invoke  CommDlgExtendedError
        test    eax, eax
        jz      .finish

        cmp     eax, $3002
        jne     .errbox

        cmp     [.errcnt], 0
        jne     .errbox

        mov     [.opn.lpstrInitialDir], 0
        inc     [.errcnt]
        jmp     .retry

.errbox:
        invoke  MessageBoxA, [hApplication], cOpenDialogError, NULL, MB_ICONERROR or MB_OK
        jmp     .finish

@@:
; Get file(s) path.

        movzx   esi, [.opn.nFileOffset]
        add     esi, edi
        cmp     byte [esi-1], 0                 ; if 0 then more than one file is selected.
        je      .multyfile                      ; The ebx string contains the filename.

; only one file to read:
        stdcall [.ptrCallBack], edi, [.lParam]
        jmp     .finish

.multyfile:
        stdcall StrNew
        mov     ebx, eax

.fileloop:
        cmp     byte [esi], 0
        je      .endfiles

        stdcall StrCopy, ebx, edi               ; path
        stdcall StrCat, ebx, cBackSlash         ; '\'
        stdcall StrCat, ebx, esi                ; path + name

        stdcall [.ptrCallBack], ebx, [.lParam]

.findend:
        lodsb
        test    al,al
        jnz     .findend
        jmp     .fileloop

.endfiles:
        stdcall StrDel, ebx

.finish:
;Free filename buffer
        stdcall FreeMem, edi
        return
endp



proc SaveFileDialog, .ptrTitle, .ptrFilter, .hFileName
.opn OPENFILENAME
begin
; Allocate memory for filename buffer.
        push      esi edi ebx
        stdcall   StrNew
        mov       ebx, eax
        stdcall   StrSetCapacity, ebx, 1024
        test      eax, eax
        jz        .finish

        mov       edi, eax
        stdcall   StrCopy, ebx, [.hFileName]

        xor     edx, edx
        mov     [.opn.lStructSize], sizeof.OPENFILENAME
        mov     eax, [hApplication]
        mov     [.opn.hwndOwner], eax
        mov     [.opn.hInstance], edx
        mov     eax, [.ptrFilter]
        mov     [.opn.lpstrFilter], eax
        mov     [.opn.lpstrCustomFilter], edx
        mov     [.opn.nFilterIndex], 1
        mov     [.opn.lpstrFile], edi
        mov     [.opn.nMaxFile], 1024
        mov     [.opn.lpstrFileTitle], edx
        mov     [.opn.nMaxFileTitle], edx
        mov     [.opn.lpstrInitialDir], edx
        mov     eax, [.ptrTitle]
        mov     [.opn.lpstrTitle], eax
        mov     [.opn.Flags], OFN_EXPLORER or           \
                              OFN_HIDEREADONLY or       \
                              OFN_NOREADONLYRETURN or   \
                              OFN_NOCHANGEDIR
        mov     [.opn.lpstrDefExt], edx
        mov     [.opn.lCustData], edx
        mov     [.opn.lpfnHook], edx

        lea     eax, [.opn]
        invoke  GetSaveFileNameA, eax
        test    eax, eax
        jz      .finish

        stdcall StrFixLen, ebx
        mov     eax, ebx
.finish:
        pop     ebx edi esi
        return
endp



proc SelectPathDialog, .hOwner, .hTitle, .hPath
.bi BROWSEINFO
.str rb 256
.path rb 1024
.IMalloc dd ?
begin
        push    esi edi
        mov     eax, [.hOwner]
        lea     esi, [.str]
        mov     [.bi.hwndOwner], eax
        mov     [.bi.pidlRoot], 0

        mov     [.bi.pszDisplayName], esi
        stdcall StrPtr, [.hTitle]
        mov     [.bi.lpszTitle], eax
        mov     [.bi.ulFlags], BIF_RETURNONLYFSDIRS
        mov     [.bi.lpfn], _BrowseFolderProc
        lea     eax, [.hOwner]
        mov     [.bi.lParam], eax

        lea     eax, [.bi]
        invoke  SHBrowseForFolder, eax
        test    eax, eax
        jz      .finish

        push    eax

        lea     edi, [.path]
        xor     esi, esi
        invoke  SHGetPathFromIDList, eax, edi
        test    eax, eax
        jz      .memfree

        stdcall StrDup, edi
        mov     esi, eax

.memfree:
        invoke  CoTaskMemFree ; From the stack
        mov     eax, esi

.finish:
        pop     edi esi
        return
endp


proc _BrowseFolderProc, .hwnd, .wmsg, .lparam, .lpData
begin
        cmp     [.wmsg], BFFM_INITIALIZED
        jne     .finish

        mov     esi, [.lpData] ; This is pointer to arguments of SelectPathDialog
        stdcall StrPtr, [esi+4]
        invoke  SendMessageA, [.hwnd], WM_SETTEXT, 0, eax
        stdcall StrPtr, [esi+8]
        invoke  SendMessageA, [.hwnd], BFFM_SETSELECTION, TRUE, eax

.finish:
        xor     eax, eax
        return
endp